#include <LCAPIIOAcc.h>
#include <LC3GlobalsDMA.h>
#include <LC3GV.h>
#include <string.h>

typedef uint32_t LcIOAccCopyBitOffs;
typedef uint8_t LcIOAccCopyBitSize;

typedef struct
{
  uint32_t bitOffset;
  uint8_t bitSize;
  void* variable;
} LcIOAccCopyEntry;

typedef struct
{
  uint32_t copyEntryCnt;
  LcIOAccCopyEntry const* copyEntries;
} LcIOAccCopyTable;

typedef uint32_t LcIOAccTableId;

typedef struct
{
  LcIOAccServiceId serviceId;
  uint32_t additionalIds[LC_IOACC_ADDITIONAL_IDS_SIZE];
  LcIOAccCopyTable *inputCopyTable;
  uint32_t inputBitSize;
  LcIOAccCopyTable *outputCopyTable;
  uint32_t outputBitSize;
  LcIOAccCopyTable *memoryCopyTable;
  uint32_t memoryBitSize;
  LcIOAccTableId tableId;
} LcIOAccTableDescriptor;

static LcIOAccCopyEntry ioCopyEntries_1_INPUT[] = 
{
  {0, 1, &(g_LC3_GVInstanceData.LC_VD_I_1)},
};

static LcIOAccCopyTable ioCopyTable_1_INPUT = {1, ioCopyEntries_1_INPUT};

static LcIOAccCopyEntry ioCopyEntries_1_OUTPUT[] = 
{
  {560, 1, &(g_LC3_GVInstanceData.LC_VD_O_1)},
  {561, 1, &(g_LC3_GVInstanceData.LC_VD_O_2)},
  {562, 1, &(g_LC3_GVInstanceData.LC_VD_O_3)},
  {563, 1, &(g_LC3_GVInstanceData.LC_VD_O_4)},
  {564, 1, &(g_LC3_GVInstanceData.LC_VD_O_5)},
  {565, 1, &(g_LC3_GVInstanceData.LC_VD_O_6)},
  {566, 1, &(g_LC3_GVInstanceData.LC_VD_O_7)},
  {567, 1, &(g_LC3_GVInstanceData.LC_VD_O_8)},
  {568, 1, &(g_LC3_GVInstanceData.LC_VD_O_9)},
  {569, 1, &(g_LC3_GVInstanceData.LC_VD_O_10)},
  {570, 1, &(g_LC3_GVInstanceData.LC_VD_O_11)},
  {571, 1, &(g_LC3_GVInstanceData.LC_VD_O_12)},
  {572, 1, &(g_LC3_GVInstanceData.LC_VD_O_13)},
  {573, 1, &(g_LC3_GVInstanceData.LC_VD_O_14)},
};

static LcIOAccCopyTable ioCopyTable_1_OUTPUT = {14, ioCopyEntries_1_OUTPUT};

static LcIOAccCopyTable ioCopyTable_1_MEMORY = {0, NULL};

#define BITS_READ(src,offs,cnt)       ((src) >> (offs) & 0xff >> (8-cnt))
#define BITS_WRITE(dst,offs,cnt,value)    \
  do{                                     \
    (dst) &= ~(BITS_READ(0xff,0,cnt) << (offs)); \
    (dst) |= (value) << (offs); \
  } while (0)

#define ADD_BIT_OFFS(ptr,offs,delta) \
  do{ \
    (offs) += (delta); \
    (ptr) += (offs) / 8; \
    (offs) %= 8; \
  } while(0)

static void copyBits(uint8_t* dst, uint8_t dstBitOffs, uint8_t const* src, uint8_t srcBitOffs, uint8_t bitCnt)
{
  uint8_t bitCntCopy;
  uint8_t bitCntSrc;
  uint8_t bitCntDst;
  uint8_t copyValue;

  while (bitCnt)
  {
    bitCntCopy = bitCnt;
    bitCntSrc = 8 - srcBitOffs;
    if(bitCntSrc < bitCntCopy)
    {
      bitCntCopy = bitCntSrc;
    }
    bitCntDst = 8 - dstBitOffs;
    if(bitCntDst < bitCntCopy)
    {
      bitCntCopy = bitCntDst;
    }
    copyValue = BITS_READ(*src, srcBitOffs, bitCntCopy);
    BITS_WRITE(*dst, dstBitOffs, bitCntCopy, copyValue);
    ADD_BIT_OFFS(dst, dstBitOffs, bitCntCopy);
    ADD_BIT_OFFS(src, srcBitOffs, bitCntCopy);
    bitCnt -= bitCntCopy;
  }
}

static void copyIOsToVariables(LcIOAccCopyTable const* copyTable, uint8_t const* ioSegment)
{
  LcIOAccCopyEntry const* copyEntry;
  LcIOAccCopyEntry const* copyEntryEnd = copyTable->copyEntries + copyTable->copyEntryCnt;

  for (copyEntry = copyTable->copyEntries; copyEntry < copyEntryEnd; copyEntry++)
  {
    copyBits(copyEntry->variable, 0, ioSegment+copyEntry->bitOffset/8, copyEntry->bitOffset % 8,copyEntry->bitSize);
  }
}

static void copyIOsFromVariables(LcIOAccCopyTable const* copyTable, uint8_t* ioSegment)
{
  LcIOAccCopyEntry const* copyEntry;
  LcIOAccCopyEntry const* copyEntryEnd = copyTable->copyEntries + copyTable->copyEntryCnt;

  for (copyEntry = copyTable->copyEntries; copyEntry < copyEntryEnd; copyEntry++)
  {
    copyBits(ioSegment + copyEntry->bitOffset / 8, copyEntry->bitOffset % 8, copyEntry->variable, 0, copyEntry->bitSize);
  }
}

static LcIOAccTableDescriptor ioTableDescriptors[1] = 
{
  {1, {0, 0}, &ioCopyTable_1_INPUT, 1, &ioCopyTable_1_OUTPUT, 574, &ioCopyTable_1_MEMORY, 0, 1},
};

static LcIOAccTableDescriptor *getTableDescriptor(LcIOAccServiceId serviceId, LcIOAccAdditionalData *additionalData)
{
  int ioTableIndex = 0;

  while (ioTableIndex < 1)
  {
    if (ioTableDescriptors[ioTableIndex].serviceId == serviceId)
    {
      if (additionalData == NULL)
      {
        return &ioTableDescriptors[ioTableIndex];
      }
      else if (additionalData->dataVersion == LC_IOACC_ADDITIONAL_DATA_VERSION(1)
          && additionalData->additionalIds[0] == ioTableDescriptors[ioTableIndex].additionalIds[0]
          && additionalData->additionalIds[1] == ioTableDescriptors[ioTableIndex].additionalIds[1])
      {
        return &ioTableDescriptors[ioTableIndex];
      }
    }
    ioTableIndex++;
  }

  return NULL;
}

LcIOAccCopyRc lc_ioacc_copyInputData(
  LcIOAccServiceId serviceId, uint8_t const* ioInputData, LcIOAccSegmentSize inputBitSize, LcIOAccAdditionalData *additionalData)
{
  LcIOAccTableDescriptor *tableDescriptor;

  tableDescriptor = getTableDescriptor(serviceId, additionalData);
  if (tableDescriptor == NULL)
  {
    return LC_IOACC_COPY_RC_INV_SERVICE_ID;
  }
  else if (inputBitSize < tableDescriptor->inputBitSize)
  {
    return LC_IOACC_COPY_RC_SEGMENT_SIZE_ERR;
  }
  else
  {
    copyIOsToVariables(tableDescriptor->inputCopyTable, ioInputData);
    return LC_IOACC_COPY_RC_OK;
  }
}

LcIOAccCopyRc lc_ioacc_copyOutputData(
  LcIOAccServiceId serviceId, uint8_t* ioOutputData, LcIOAccSegmentSize outputBitSize, LcIOAccAdditionalData* additionalData)
{
  LcIOAccTableDescriptor *tableDescriptor;

  tableDescriptor = getTableDescriptor(serviceId, additionalData);
  if (tableDescriptor == NULL)
  {
    return LC_IOACC_COPY_RC_INV_SERVICE_ID;
  }
  else if (outputBitSize < tableDescriptor->outputBitSize)
  {
    return LC_IOACC_COPY_RC_SEGMENT_SIZE_ERR;
  }
  else
  {
    copyIOsFromVariables(tableDescriptor->outputCopyTable, ioOutputData);
    return LC_IOACC_COPY_RC_OK;
  }
}

LcIOAccCopyRc lc_ioacc_copyMemoryDataToVariables(
  LcIOAccServiceId serviceId, uint8_t const* ioInputData, LcIOAccSegmentSize inputBitSize, LcIOAccAdditionalData *additionalData)
{
  LcIOAccTableDescriptor *tableDescriptor;

  tableDescriptor = getTableDescriptor(serviceId, additionalData);
  if (tableDescriptor == NULL)
  {
    return LC_IOACC_COPY_RC_INV_SERVICE_ID;
  }
  else if (inputBitSize < tableDescriptor->memoryBitSize)
  {
    return LC_IOACC_COPY_RC_SEGMENT_SIZE_ERR;
  }
  else
  {
    copyIOsToVariables(tableDescriptor->memoryCopyTable, ioInputData);
    return LC_IOACC_COPY_RC_OK;
  }
}

LcIOAccCopyRc lc_ioacc_copyMemoryDataFromVariables(
  LcIOAccServiceId serviceId, uint8_t* ioOutputData, LcIOAccSegmentSize outputBitSize, LcIOAccAdditionalData* additionalData)
{
  LcIOAccTableDescriptor *tableDescriptor;

  tableDescriptor = getTableDescriptor(serviceId, additionalData);
  if (tableDescriptor == NULL)
  {
    return LC_IOACC_COPY_RC_INV_SERVICE_ID;
  }
  else if (outputBitSize < tableDescriptor->memoryBitSize)
  {
    return LC_IOACC_COPY_RC_SEGMENT_SIZE_ERR;
  }
  else
  {
    copyIOsFromVariables(tableDescriptor->memoryCopyTable, ioOutputData);
    return LC_IOACC_COPY_RC_OK;
  }
}

extern LCoplck_TASK LC_vg_Task_DEFAULTTASK;

#include <ioAccAdditionalData_RevolutionPi_1_Def.inc>

static LcIOAccServiceInfo ioServiceInfos[] = 
{
  {"RevolutionPi",1,&LC_vg_Task_DEFAULTTASK,&ioAccAdditionalData_RevolutionPi_1,1,574,0},
};

LcIOAccInfoRc lc_ioacc_getServiceInfoByName(char const* const serviceName, LcIOAccServiceInfo* serviceInfo)
{
  LcIOAccServiceInfo* serviceInfoIter;
  for(serviceInfoIter = ioServiceInfos; serviceInfoIter < ioServiceInfos + 1; serviceInfoIter++)
  {
    if(strcmp(serviceInfoIter->serviceName, serviceName) == 0)
    {
      *serviceInfo = *serviceInfoIter;
      return LC_IOACC_INFO_RC_OK;
    }
  }
  return LC_IOACC_INFO_RC_SERVICE_NOT_FOUND;
}

